-- diskutil.e
-- disk utilities
-- J. Craig Gilbert
-- cgilbert@cennet.mc.peachnet.edu
-- 3/01/98
-- CREDIT(S):
--    floppies() is derived from David Cuny's routine of the same name
--    drive_list() is a direct copy of David Cuny's drive_list()
include wildcard.e
include machine.e

-- local includes
include mach.e

sequence mregs
-----------------------------------------------------------------------------
global function current_drive()
-- returns the letter of the current default drive
   mregs = repeat(0,10)
   mregs[REG_AX] = #1900
   mregs = dos_interrupt(#21,mregs)
   return (and_bits(mregs[REG_AX],#FF) + 'A')
end function
-----------------------------------------------------------------------------
global procedure set_drive(integer letter)
-- sets the current default drive
-- NOTE: even if the drive is not ready, this will set it as the default
-- Will not set the drive if the drive does not exist
   letter = upper(letter)
   if letter < 'A' or letter > 'Z' then
      return
   end if
   mregs = repeat(0,10)
   mregs[REG_AX] = #E00
   mregs[REG_DX] = letter - 'A'
   mregs = dos_interrupt(#21, mregs)
end procedure
-----------------------------------------------------------------------------
global function drive_exist(integer letter)
-- Returns 1 if the drive exists, 0 otherwise
integer current, prev, drive_status

   letter = upper(letter)
   if letter < 'A' or letter > 'Z' then
      return 0
   end if

   prev = current_drive()

   set_drive(letter)
   current = current_drive()
   if current = letter then
      drive_status = 1
   else
      drive_status = 0
   end if

   set_drive(prev)

   return drive_status
end function
-----------------------------------------------------------------------------
procedure force_drive_access(integer letter, atom error_ptr)
-- attempt disk access on the given drive
-- the byte at error_ptr should be preset to #FF
atom f
integer file_was_there
sequence fname

   file_was_there = 0
   fname = {letter} & ":\\zzwgarbl"

   -- check by read first
   f = open(fname, "r")
   if f != -1 then
      file_was_there = 1
      close(f)
   end if

   if peek(error_ptr) != #FF then
      return
   end if

   -- now check by write
   f = open(fname,"a")
   if f != -1 then
      close(f)
      if not file_was_there then
         system("del " & fname,2)
      end if
   end if

   return
end procedure
------------------------------------------
global constant   DRIVE_READY = -1,   DRIVE_NOT_EXIST = -2,
                  DRIVE_WRITE_PROTECTED = 0,
                  DRIVE_NOT_READY = 2
global function check_drive(integer letter)
-- This function replaces the critical error handler with it's own,
-- attempts to access drive 'letter', then restores the previous
-- critical error handler.
-- Returns -1 if drive is ready for all kinds of access
--         -2 if drive does not exist
--         the critical error code from DOS otherwise
--
-- The codes:
--        00  write protect error <-------+
--        01  unknown unit                |- most likely to see
--        02  drive not ready  <----------+
--        03  unknown command
--        04  data error (bad CRC)
--        05  bad request structure length
--        06  seek error
--        07  unknown media type
--        08  sector not found
--        09  printer out of paper
--        0A  write fault
--        0B  read fault
--        0C  general failure <----------- somewhat likely
atom new_hand_ptr, errcode_ptr
integer critical_code
sequence new_hand, data_segment

   letter = upper(letter)

   if not drive_exist(letter) then
      return DRIVE_NOT_EXIST
   end if

   data_segment = int_to_bytes( get_seg(1) )

   -- create error code location, preloaded w/ #ff (code not used by DOS)
   errcode_ptr = mem_locked(2)   -- high byte is ignored but needs space
   poke(errcode_ptr, #FF)

   -- construct new handler
   new_hand = {   -- save stuff
                  #1E,  -- push ds
                  #53,  -- push ebx
                  -- load data segment
                  #BB} & data_segment & -- mov ebx, data segment value
                { #53,  -- push ebx
                  #1F} &  -- pop ds

                  -- do your thing
                { #89, #3D } & int_to_bytes(errcode_ptr) & -- mov mem, di

                  -- acknowledge interrupt w/ ignore (zero in al)
                { #32, #C0 } & -- xor eal,eal

                  -- restore stuff
                { #5B,  -- pop ebx
                  #1F,  -- pop ds

                  #CF}  -- iretd

   -- set up new handler
   new_hand_ptr = install_handler(new_hand, #24)

   -- access the drive
   force_drive_access(letter, errcode_ptr)

   -- reassign INT 24h to previous critical error handler
   reset_handler(#24)

   -- get err code
   critical_code = peek(errcode_ptr)

   free(errcode_ptr)

   if critical_code = #FF then
      return DRIVE_READY
   else
      return critical_code
   end if

end function
-----------------------------------------------------------------------------
global function floppies()
--  Based on David Cuny's function of the same name
--  Peek() at the BIOS Equipment Word to return a list of floppy
--  drives installed on the system.
--
-- CREDIT(S)
--  Larry Stone, based on a routine published in MicroHelp's BUG
--  Newsletter, 1/1/90.
integer count
sequence list

    -- get count of floppies
    count = floor( peek( #410 ) / 64 + 1 )

    -- clear the list
    list = repeat(0, count)

    -- set floppy letters
    for drive = 1 to count do
        list[drive] =  'A' + drive - 1
    end for

return list
end function
-----------------------------------------------------------------------------
global function hard_drives()
-- return list of hard drives
sequence flops, list
integer first_hd

   -- get floppy list
   flops = floppies()

   if length( flops ) <= 1 then
      first_hd = 'C'
   else
      first_hd = flops[length(flops)] + 1
   end if

   -- check possible hard drives
   list = {}
   for cur = first_hd to 'Z' do
     if drive_exist(cur) then
        list = list & {cur}
     end if
   end for

return list
end function
----------------------------------------------------------------------------
global function drive_list()
-- original author David Cuny
   return (floppies() & hard_drives())
end function
-----------------------------------------------------------------------------
global function disk_free_space(integer letter)
-- get current free space on specified drive
-- Returns: free space in bytes
-- if the specified drive does not exist, you will get -1
-- Note: You will get a critical error if the drive is not ready; see
-- the func check_drive().
integer prev, bytes_per_sector, sectors_per_cluster, free_clusters
atom free_space

   if not drive_exist(letter) then
      return -1
   end if
   
   prev = current_drive()
   set_drive(letter)

   mregs = repeat(0,10)
   mregs[REG_AX] = #3600
   mregs = dos_interrupt(#21,mregs)
   if mregs[REG_AX] = #FFFF then
      free_space = 0
   else
      bytes_per_sector = mregs[REG_CX]
      sectors_per_cluster = mregs[REG_AX]
      free_clusters = mregs[REG_BX]
      free_space = bytes_per_sector * sectors_per_cluster * free_clusters
   end if

   set_drive(prev)
return free_space
end function
-----------------------------------------------------------------------------
function safe_dta()
-- return a memory address pointing to 64 bytes allocated in low memory,to be
-- used for a DTA.
-- I'm not really sure of this, but according to my source,
-- a DTA should not be situated so that it crosses a 64K segment
-- boundary.  I *think* this function will return a safe address,
-- although I don't know if this restriction even applies in protected mode.
-- It's worked so far, but use at your own risk . . .
atom new1, new2, closest_greater

   new1 = allocate_low(64)

   closest_greater = (floor(new1/(64*1024)) + 1) * 64*1024
   if (closest_greater - new1) <= 64 then
      new2 = safe_dta()
      free(new1)
      return new2
   else
      return new1
   end if
end function
------------------------
function current_dta()
-- returns {segment,offset} of current DTA
   mregs = repeat(0,10)
   mregs[REG_AX] = #2F00
   mregs = dos_interrupt(#21,mregs)
   return {mregs[REG_ES],mregs[REG_BX]}
end function
------------------------
procedure set_dta(sequence d)
-- set DTA to address specified by the sequence d:
-- {segment, offset}
   mregs = repeat(0,10)
   mregs[REG_AX] = #1A00
   mregs[REG_DS] = d[1]
   mregs[REG_DX] = d[2]
   mregs = dos_interrupt(#21, mregs)
end procedure
------------------------
procedure set_temp_dta(atom new_dta_ptr)
-- set up temporary dta space
atom new_dta_off, new_dta_seg
   new_dta_off = remainder(new_dta_ptr,16)
   new_dta_seg = floor(new_dta_ptr/16)
   set_dta({new_dta_seg, new_dta_off})
end procedure
------------------------
global function set_vol_label(integer letter, sequence new_label)
-- Sets volume label of disk specified by 'letter'.
-- Returns 1 for success, 0 for failure.
-- Note 1: You will get a critical error if the drive is not ready; see
-- the func check_drive().
sequence prev_dta
integer prev_drive, failed
atom new_dta_ptr, xfcb_off, xfcb_seg, xfcb_ptr

   if not drive_exist(letter) then
      return 0
   end if

   new_label = sprintf("%-11.11s",{new_label})

   prev_drive = current_drive()
   set_drive(letter)

   prev_dta = current_dta()
   new_dta_ptr = safe_dta()
   if new_dta_ptr = 0 then
      return 0
   end if
   set_temp_dta(new_dta_ptr)

   -- as far as I can tell, it should be 52 bytes long, but just in
   -- case I'm creating it 64 bytes in size; my docs are vague on that.
   xfcb_ptr = allocate_low(64)
   if xfcb_ptr = 0 then
      free_low(new_dta_ptr)
      return 0
   end if

   mem_set(xfcb_ptr,0,64) -- blank everything
                          -- this blanking also serves the purpose of setting
                          -- the drive field of xfcb to 'default'
   mem_set(xfcb_ptr+8, '?', 11) -- set name to all wildcard char '?'

   poke(xfcb_ptr,#FF)    -- indicate that this is an xfcb
   poke(xfcb_ptr+6, #08) -- set attribute to 'label'

   
   -- search for first file matching xfcb (any file w/ label attribute)
   xfcb_seg = floor(xfcb_ptr/16)
   xfcb_off = remainder(xfcb_ptr,16)
   mregs = repeat(0,10)
   mregs[REG_AX] = #1100
   mregs[REG_DX] = xfcb_off
   mregs[REG_DS] = xfcb_seg
   mregs = dos_interrupt(#21, mregs)
   mregs[REG_AX] = and_bits(mregs[REG_AX], #FF)

   if mregs[REG_AX] = 0 then -- label exists, rename it
      poke(xfcb_ptr+24,new_label)
      mregs[REG_AX] = #1700
      mregs[REG_DX] = xfcb_off
      mregs[REG_DS] = xfcb_seg
      mregs = dos_interrupt(#21, mregs)
      failed = and_bits(mregs[REG_AX], #FF)
   else
      -- create new label
      poke(xfcb_ptr+8,new_label)
      mregs[REG_AX] = #1600
      mregs[REG_DX] = xfcb_off
      mregs[REG_DS] = xfcb_seg
      mregs = dos_interrupt(#21, mregs)
      failed = and_bits(mregs[REG_AX], #FF)
      -- close the label file
      mregs[REG_AX] = #1000
      mregs[REG_DX] = xfcb_off
      mregs[REG_DS] = xfcb_seg
      mregs = dos_interrupt(#21, mregs)
   end if

   set_dta(prev_dta)
   set_drive(prev_drive)

   free_low(new_dta_ptr)
   free_low(xfcb_ptr)

   if not failed then
      return 1
   else
      return 0
   end if
end function
------------------------------------------------------------------------------
global function get_disk_id(integer drive)
-- get serial number, volume label, file system
-- requires DOS 4.0 or greater
-- returns sequence {{serno}, {label}, {filesys}}
-- if error returns 0
atom buffer_ptr
sequence label, serno, filesys

   buffer_ptr = allocate_low(25)
   if buffer_ptr = -1 then
      return 0
   else
      lock_memory(buffer_ptr, 25)
   end if
   
   poke(buffer_ptr, repeat(0,25) )

   mregs = repeat(0,10)
   mregs[REG_AX] = #6900
   mregs[REG_BX] = upper(drive) - '@'
   mregs[REG_DS] = floor(buffer_ptr/16)
   mregs[REG_DX] = remainder(buffer_ptr,16)

   mregs = dos_interrupt(#21,mregs)
   if and_bits(mregs[REG_FLAGS], 1) = 0 then
      serno = sprintf("%-8.8x",{bytes_to_int(peek({buffer_ptr + 2, 4 }))})
      label = peek({buffer_ptr+6,11})
      filesys = peek({buffer_ptr+#11,8})

      free_low(buffer_ptr)
      
      return {serno, label, filesys}
   else
      free_low(buffer_ptr)
      return 0
   end if
end function
------------------------------------------------------------------------------
global function set_disk_serial(integer drive, atom new_serial)
-- set serial number
-- requires DOS 4.0 or greater
-- returns 1 success
--         0 failure
atom buffer_ptr
object disk_info

   disk_info = get_disk_id(drive)
   if not sequence(disk_info) then
      return 0
   else
      disk_info[1] = new_serial
   end if

   buffer_ptr = allocate_low(25)
   if buffer_ptr = -1 then
      return 0
   else
      lock_memory(buffer_ptr, 25)
   end if
   
   poke(buffer_ptr, {0,0} )
   poke(buffer_ptr+2, int_to_bytes(disk_info[1]) )
   poke(buffer_ptr+6, sprintf("%-11.11s",{disk_info[2]}) )
   poke(buffer_ptr+#11, sprintf("%-8.8s",{disk_info[3]}) )

   mregs = repeat(0,10)
   mregs[REG_AX] = #6901
   mregs[REG_BX] = upper(drive) - '@'
   mregs[REG_DS] = floor(buffer_ptr/16)
   mregs[REG_DX] = remainder(buffer_ptr,16)

   mregs = dos_interrupt(#21,mregs)
   free_low(buffer_ptr)
   if and_bits(mregs[REG_FLAGS], 1) = 0 then
      return 1
   else
      return 0
   end if
end function
------------------------------------------------------------------------------
-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-----------------------------------------------------------------------------

